# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add namespace history

Revision ID: 5c7a5c5a193b
Revises: c2d6da81f9ec
Create Date: 2024-02-27 13:25:07.367475

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5c7a5c5a193b"
down_revision = "c2d6da81f9ec"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "namespace_history",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("namespace_id", sa.BigInteger(), nullable=True),
        sa.Column("name", sa.String(length=2048), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("action", sa.String(length=255), nullable=False),
        sa.Column("changed_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=False),
        sa.Column("changed_by_user_id", sa.BigInteger(), nullable=True),
        sa.ForeignKeyConstraint(
            ["changed_by_user_id"],
            ["user.id"],
            name=op.f("fk__namespace_history__changed_by_user_id__user"),
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__namespace_history")),
    )
    op.create_index(op.f("ix__namespace_history__namespace_id"), "namespace_history", ["namespace_id"], unique=False)

    op.execute(
        """
        INSERT INTO namespace_history (namespace_id, name, description, action, changed_at, changed_by_user_id)
        SELECT id, name, description, 'Deleted', changed_at, changed_by_user_id
        FROM namespace
        WHERE is_deleted = TRUE
    """,
    )
    op.execute(
        """
        DELETE FROM namespace WHERE is_deleted = TRUE
    """,
    )
    op.drop_column("namespace", "is_deleted")


def downgrade() -> None:
    op.add_column(
        "namespace",
        sa.Column("is_deleted", sa.BOOLEAN(), autoincrement=False, nullable=False, server_default="FALSE"),
    )

    op.execute(
        """
        UPDATE namespace SET is_deleted = TRUE
        WHERE id IN (
            SELECT namespace_id FROM namespace_history WHERE action = 'Deleted'
        )
    """,
    )
    op.drop_index(op.f("ix__namespace_history__namespace_id"), table_name="namespace_history")
    op.drop_table("namespace_history")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

"""Add user table

Revision ID: b91de692624e
Revises:
Create Date: 2023-10-09 18:32:01.037500

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b91de692624e"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("username", sa.String(length=256), nullable=False),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__user")),
    )
    op.create_index(op.f("ix__user__username"), "user", ["username"], unique=True)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix__user__username"), table_name="user")
    op.drop_table("user")
    # ### end Alembic commands ###



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap

from pydantic import BaseModel, Field


class DatabaseSettings(BaseModel):
    """Database connection settings.

    .. note::

        You can pass here any extra option supported by
        `SQLAlchemy Engine class <https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine>`_,
        even if it is not mentioned in documentation.

    Examples
    --------

    .. code-block:: bash

        HORIZON__DATABASE__URL=postgresql+asyncpg://postgres:postgres@localhost:5432/horizon
        # custom option passed directly to engine factory
        HORIZON__DATABASE__POOL_PRE_PING=True
    """

    url: str = Field(
        description=textwrap.dedent(
            """
            Database connection URL.

            See `SQLAlchemy documentation <https://docs.sqlalchemy.org/en/20/core/engines.html#backend-specific-urls>`_

            .. warning:

                Only async drivers are supported, e.g. ``asyncpg``
            """,
        ),
    )

    class Config:
        extra = "allow"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import Any


class ApplicationError(ABC, Exception):
    """Base class for all exceptions raised by Horizon."""

    @property
    @abstractmethod
    def message(self) -> str:
        """Message string"""
        ...

    @property
    @abstractmethod
    def details(self) -> Any:
        """Details related to specific error"""
        ...

    def __str__(self) -> str:
        return self.message



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import APIRouter, Depends
from fastapi.security import OAuth2PasswordRequestForm
from typing_extensions import Annotated

from horizon.backend.dependencies.stub import Stub
from horizon.backend.providers.auth import AuthProvider
from horizon.commons.errors import get_error_responses
from horizon.commons.errors.schemas import InvalidRequestSchema, NotAuthorizedSchema
from horizon.commons.schemas.v1 import AuthTokenResponseV1

router = APIRouter(prefix="/auth", tags=["Auth"])


@router.post(
    "/token",
    summary="Get access token",
    responses=get_error_responses(include={NotAuthorizedSchema, InvalidRequestSchema}),
)
async def login(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()],
    auth_provider: Annotated[AuthProvider, Depends(Stub(AuthProvider))],
) -> AuthTokenResponseV1:
    token = await auth_provider.get_token(
        grant_type=form_data.grant_type,
        login=form_data.username,
        password=form_data.password,
        scopes=form_data.scopes,
        client_id=form_data.client_id,
        client_secret=form_data.client_secret,
    )
    return AuthTokenResponseV1.parse_obj(token)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add namespace table

Revision ID: e29b66594970
Revises: b91de692624e
Create Date: 2023-10-13 12:24:06.942579

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e29b66594970"
down_revision = "b91de692624e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "namespace",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column(
            "changed_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            server_onupdate=sa.text("now()"),  # type: ignore[arg-type]
            nullable=False,
        ),
        sa.Column("changed_by_user_id", sa.BigInteger(), nullable=True),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["changed_by_user_id"],
            ["user.id"],
            name=op.f("fk__namespace__changed_by_user_id__user"),
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__namespace")),
    )
    op.create_index(op.f("ix__namespace__name"), "namespace", ["name"], unique=True)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix__namespace__name"), table_name="namespace")
    op.drop_table("namespace")
    # ### end Alembic commands ###



#!/bin/env python3

# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import sys
from pathlib import Path

from alembic.config import CommandLine

here = Path(__file__).resolve()
config_path = here.parent / "alembic.ini"


def main(prog_name: str | None = None, args: list[str] | None = None):
    """Run alembic and pass the command line arguments to it."""
    if args is None:
        args = sys.argv.copy()
        prog_name = args.pop(0)

    if not prog_name:
        prog_name = os.fspath(here)

    # prepend config path before command line arguments
    args = args.copy()
    args.insert(0, "-c")
    args.insert(1, os.fspath(config_path))

    # call alembic
    cmd = CommandLine(prog=prog_name)
    cmd.main(argv=args)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from enum import IntEnum

from sqlalchemy import BigInteger, ForeignKey, String, Text
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import Mapped, mapped_column, relationship

from horizon.backend.db.mixins.changed_by import ChangedByMixin
from horizon.backend.db.models.base import Base


class NamespaceUserRoleInt(IntEnum):
    GUEST = 0
    DEVELOPER = 1
    MAINTAINER = 2
    OWNER = 3
    SUPERADMIN = 4


class Namespace(Base, ChangedByMixin):
    __tablename__ = "namespace"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    name: Mapped[str] = mapped_column(
        String(256),  # noqa: WPS432
        nullable=False,
        unique=True,
        index=True,
    )
    description: Mapped[str] = mapped_column(
        Text(),
        nullable=False,
    )
    owner_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("user.id", ondelete="RESTRICT"),
        nullable=False,
    )
    owner = relationship("User", foreign_keys=[owner_id])
    owned_by: AssociationProxy[str] = association_proxy("owner", "username")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator

from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker


def sessionmaker(engine: AsyncEngine) -> async_sessionmaker[AsyncSession]:
    return async_sessionmaker(
        bind=engine,
        class_=AsyncSession,
        expire_on_commit=False,
    )


def create_session_factory(engine: AsyncEngine):
    Session = sessionmaker(engine)  # noqa: N806

    async def wrapper() -> AsyncGenerator[AsyncSession, None]:
        async with Session.begin() as session:
            yield session

    return wrapper



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from horizon.commons.schemas.v1.auth import AuthTokenResponseV1
from horizon.commons.schemas.v1.hwm import (
    HWMBulkCopyRequestV1,
    HWMBulkDeleteRequestV1,
    HWMCreateRequestV1,
    HWMListResponseV1,
    HWMPaginateQueryV1,
    HWMResponseV1,
    HWMUpdateRequestV1,
)
from horizon.commons.schemas.v1.hwm_history import (
    HWMHistoryPaginateQueryV1,
    HWMHistoryResponseV1,
)
from horizon.commons.schemas.v1.namespace import (
    NamespaceCreateRequestV1,
    NamespacePaginateQueryV1,
    NamespaceResponseV1,
    NamespaceUpdateRequestV1,
    NamespaceUserRole,
)
from horizon.commons.schemas.v1.namespace_history import (
    NamespaceHistoryPaginateQueryV1,
    NamespaceHistoryResponseV1,
)
from horizon.commons.schemas.v1.pagination import (
    PageMetaResponseV1,
    PageResponseV1,
    PaginateQueryV1,
)
from horizon.commons.schemas.v1.permission import (
    PermissionResponseItemV1,
    PermissionsResponseV1,
    PermissionsUpdateRequestV1,
    PermissionUpdateRequestItemV1,
)
from horizon.commons.schemas.v1.user import UserResponseV1, UserResponseV1WithAdmin

__all__ = [
    "AuthTokenResponseV1",
    "HWMCreateRequestV1",
    "HWMPaginateQueryV1",
    "HWMResponseV1",
    "HWMListResponseV1",
    "HWMUpdateRequestV1",
    "HWMHistoryPaginateQueryV1",
    "HWMHistoryResponseV1",
    "HWMBulkCopyRequestV1",
    "HWMBulkDeleteRequestV1",
    "NamespaceCreateRequestV1",
    "NamespacePaginateQueryV1",
    "NamespaceResponseV1",
    "NamespaceUpdateRequestV1",
    "NamespaceHistoryPaginateQueryV1",
    "NamespaceHistoryResponseV1",
    "PageMetaResponseV1",
    "PageResponseV1",
    "PaginateQueryV1",
    "PermissionUpdateRequestItemV1",
    "PermissionResponseItemV1",
    "PermissionsResponseV1",
    "PermissionsUpdateRequestV1",
    "UserResponseV1",
    "UserResponseV1WithAdmin",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import http
from typing import Any

from typing_extensions import Literal

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import register_error_response
from horizon.commons.exceptions.auth import AuthorizationError


@register_error_response(
    exception=AuthorizationError,
    status=http.HTTPStatus.UNAUTHORIZED,
)
class NotAuthorizedSchema(BaseErrorSchema):
    code: Literal["unauthorized"] = "unauthorized"
    details: Any = None

    def to_exception(self) -> AuthorizationError:
        return AuthorizationError(
            message=self.message,
            details=self.details,
        )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="pydantic-orm"

from fastapi import APIRouter, Depends, HTTPException, status
from typing_extensions import Annotated

from horizon.backend.db.models import NamespaceUserRoleInt, User
from horizon.backend.services import UnitOfWork, current_user
from horizon.commons.errors import get_error_responses
from horizon.commons.exceptions import BadRequestError
from horizon.commons.schemas.v1 import (
    NamespaceCreateRequestV1,
    NamespacePaginateQueryV1,
    NamespaceResponseV1,
    NamespaceUpdateRequestV1,
    NamespaceUserRole,
    PageResponseV1,
    PermissionResponseItemV1,
    PermissionsResponseV1,
    PermissionsUpdateRequestV1,
)

router = APIRouter(prefix="/namespaces", tags=["Namespace"], responses=get_error_responses())


@router.get(
    "/",
    summary="Paginage namespaces",
    dependencies=[Depends(current_user)],
)
async def paginate_namespaces(
    pagination_args: Annotated[NamespacePaginateQueryV1, Depends()],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> PageResponseV1[NamespaceResponseV1]:
    pagination = await unit_of_work.namespace.paginate(**pagination_args.dict())
    return PageResponseV1[NamespaceResponseV1].from_pagination(pagination)


@router.get(
    "/{namespace_id}",
    summary="Get namespace",
    dependencies=[Depends(current_user)],
)
async def get_namespace(
    namespace_id: int,
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> NamespaceResponseV1:
    namespace = await unit_of_work.namespace.get(namespace_id)
    return NamespaceResponseV1.from_orm(namespace)


@router.post(
    "/",
    summary="Create namespace",
    status_code=status.HTTP_201_CREATED,
)
async def create_namespace(
    data: NamespaceCreateRequestV1,
    unit_of_work: Annotated[UnitOfWork, Depends()],
    user: Annotated[User, Depends(current_user)],
) -> NamespaceResponseV1:
    async with unit_of_work:
        namespace = await unit_of_work.namespace.create(**data.dict(), user=user)
        await unit_of_work.namespace_history.create(
            namespace_id=namespace.id,
            data=namespace.to_dict(exclude={"id"}),
        )
    return NamespaceResponseV1.from_orm(namespace)


@router.patch(
    "/{namespace_id}",
    summary="Update namespace",
)
async def update_namespace(
    namespace_id: int,
    changes: NamespaceUpdateRequestV1,
    unit_of_work: Annotated[UnitOfWork, Depends()],
    user: Annotated[User, Depends(current_user)],
) -> NamespaceResponseV1:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            namespace_id=namespace_id,
            required_role=NamespaceUserRoleInt.OWNER,
        )
        namespace = await unit_of_work.namespace.update(
            namespace_id=namespace_id,
            changes=changes.dict(exclude_defaults=True),
            user=user,
        )
        await unit_of_work.namespace_history.create(
            namespace_id=namespace.id,
            data={
                **namespace.to_dict(exclude={"id"}),
                "action": "Updated",
            },
        )
    return NamespaceResponseV1.from_orm(namespace)


@router.delete(
    "/{namespace_id}",
    summary="Delete namespace",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_namespace(
    namespace_id: int,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> None:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            namespace_id=namespace_id,
            required_role=NamespaceUserRoleInt.OWNER,
        )
        hwm_records = await unit_of_work.hwm.paginate(namespace_id=namespace_id, page=1, page_size=1)
        if hwm_records.total_count:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Cannot delete namespace because it has related HWM records.",
            )

        namespace = await unit_of_work.namespace.delete(namespace_id=namespace_id, user=user)
        await unit_of_work.namespace_history.create(
            namespace_id=namespace_id,
            data={**namespace.to_dict(exclude={"id"}), "action": "Deleted", "changed_by_user_id": user.id},
        )


@router.get("/{namespace_id}/permissions", summary="Get namespace permissions", dependencies=[Depends(current_user)])
async def get_namespace_permissions(
    namespace_id: int,
    unit_of_work: Annotated[UnitOfWork, Depends()],
    user: Annotated[User, Depends(current_user)],
) -> PermissionsResponseV1:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            namespace_id=namespace_id,
            required_role=NamespaceUserRoleInt.OWNER,
        )
        permissions_dict = await unit_of_work.namespace.get_namespace_users_permissions(namespace_id)

        permissions_response = [
            PermissionResponseItemV1(username=user.username, role=role.name) for user, role in permissions_dict.items()
        ]

    return PermissionsResponseV1(permissions=permissions_response)


@router.patch(
    "/{namespace_id}/permissions",
    summary="Update namespace permissions",
    dependencies=[Depends(current_user)],
)
async def update_namespace_permissions(  # noqa: WPS217
    namespace_id: int,
    changes: PermissionsUpdateRequestV1,
    unit_of_work: Annotated[UnitOfWork, Depends()],
    user: Annotated[User, Depends(current_user)],
) -> PermissionsResponseV1:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            namespace_id=namespace_id,
            required_role=NamespaceUserRoleInt.OWNER,
        )

        owner_change_detected = False

        for perm in changes.permissions:
            update_user = await unit_of_work.user.get_by_username(perm.username)

            # Ensure the current owner isn't changing their role without assigning a new owner
            if update_user.id == user.id and perm.role != NamespaceUserRole.OWNER:
                if not any(
                    p.role == NamespaceUserRole.OWNER for p in changes.permissions if p.username != user.username
                ):
                    raise BadRequestError("Cannot drop ownership without assigning a new owner.")

            if perm.role:
                if perm.role == NamespaceUserRole.OWNER and not owner_change_detected:
                    await unit_of_work.namespace.set_new_owner(namespace_id, update_user.id)
                    owner_change_detected = True
                else:
                    role_enum = NamespaceUserRoleInt[perm.role.upper()]
                    await unit_of_work.namespace.update_permission(namespace_id, update_user.id, role_enum)
            else:
                await unit_of_work.namespace.delete_permission(namespace_id, update_user.id)

        permissions_dict = await unit_of_work.namespace.get_namespace_users_permissions(namespace_id)
        permissions_response = [
            PermissionResponseItemV1(username=user.username, role=role.name) for user, role in permissions_dict.items()
        ]
        return PermissionsResponseV1(permissions=permissions_response)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from horizon.backend.settings.server import CORSSettings


def apply_cors_middleware(app: FastAPI, settings: CORSSettings) -> FastAPI:
    """Add CORS middleware to the application."""
    if not settings:
        return app

    app.add_middleware(
        CORSMiddleware,
        **settings.dict(exclude={"enabled"}),
    )
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from sqlalchemy import MetaData, inspect
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy_utils import generic_repr

convention = {
    "all_column_names": lambda constraint, table: "_".join(
        [column.name for column in constraint.columns.values()],
    ),
    "ix": "ix__%(table_name)s__%(all_column_names)s",
    "uq": "uq__%(table_name)s__%(all_column_names)s",
    "ck": "ck__%(table_name)s__%(constraint_name)s",
    "fk": ("fk__%(table_name)s__%(all_column_names)s__" "%(referred_table_name)s"),
    "pk": "pk__%(table_name)s",
}

horizon_metadata = MetaData(naming_convention=convention)  # type: ignore


@generic_repr
class Base(DeclarativeBase):
    metadata = horizon_metadata

    def to_dict(self, exclude: set[str] | None = None):
        exclude = exclude or set()
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs if c.key not in exclude}



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from horizon.commons.errors.schemas.already_exists import AlreadyExistsSchema
from horizon.commons.errors.schemas.bad_request import BadRequestError
from horizon.commons.errors.schemas.invalid_request import InvalidRequestSchema
from horizon.commons.errors.schemas.not_authorized import NotAuthorizedSchema
from horizon.commons.errors.schemas.not_found import NotFoundSchema
from horizon.commons.errors.schemas.permission_denied import PermissionDeniedSchema

__all__ = [
    "AlreadyExistsSchema",
    "InvalidRequestSchema",
    "NotAuthorizedSchema",
    "NotFoundSchema",
    "BadRequestError",
    "PermissionDeniedSchema",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add action to HWMHistory table

Revision ID: c2d6da81f9ec
Revises: 4bc3fffc0209
Create Date: 2024-02-22 15:47:40.155259

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "c2d6da81f9ec"
down_revision = "4bc3fffc0209"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("ALTER TABLE hwm_history DROP CONSTRAINT fk__hwm_history__hwm_id__hwm")
    op.execute("ALTER TABLE hwm_history ADD COLUMN action VARCHAR(255)")

    op.execute("UPDATE hwm_history SET action = 'Deleted' WHERE is_deleted = TRUE")

    op.execute("UPDATE hwm_history SET action = 'Updated' WHERE is_deleted = FALSE")

    op.execute(
        """
        WITH ranked_history AS (
            SELECT id, hwm_id, ROW_NUMBER() OVER(PARTITION BY hwm_id ORDER BY id) AS rank
            FROM hwm_history
        )
        UPDATE hwm_history SET action = 'Created' WHERE id IN (
            SELECT id FROM ranked_history WHERE rank = 1
        )
    """,
    )
    op.execute("DELETE FROM hwm WHERE is_deleted = TRUE")

    op.execute("ALTER TABLE hwm_history DROP COLUMN is_deleted")
    op.execute("ALTER TABLE hwm DROP COLUMN is_deleted")
    op.execute("ALTER TABLE hwm_history ALTER COLUMN action SET NOT NULL")


def downgrade() -> None:
    op.execute("ALTER TABLE hwm_history ADD COLUMN is_deleted BOOLEAN NOT NULL DEFAULT FALSE")
    op.execute("UPDATE hwm_history SET is_deleted = TRUE WHERE action = 'Deleted'")
    op.execute("ALTER TABLE hwm ADD COLUMN is_deleted BOOLEAN NOT NULL DEFAULT FALSE")
    op.execute("ALTER TABLE hwm_history DROP COLUMN action")
    op.execute(
        """
        ALTER TABLE hwm_history ADD CONSTRAINT fk__hwm_history__hwm_id__hwm
        FOREIGN KEY (hwm_id) REFERENCES hwm (id) ON DELETE CASCADE
    """,
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap
from typing import Dict, Set

from pydantic import BaseModel, Field


class MonitoringSettings(BaseModel):
    """Monitoring Settings.

    See `starlette-exporter <https://github.com/stephenhillier/starlette_exporter#options>`_ documentation.

    .. note::

        You can pass here any extra option supported by ``starlette-exporter``,
        even if it is not mentioned in documentation.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__MONITORING__ENABLED=True
        HORIZON__SERVER__MONITORING__SKIP_PATHS=["/some/path"]
        HORIZON__SERVER__MONITORING__SKIP_METHODS=["OPTIONS"]
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable middleware")
    labels: Dict[str, str] = Field(
        default_factory=dict,
        description="""Custom labels added to all metrics, e.g. ``{"instance": "production"}``""",
    )
    skip_paths: Set[str] = Field(
        default_factory=set,
        description="Custom paths should be skipped from metrics, like ``/some/endpoint``",
    )
    skip_methods: Set[str] = Field(
        default={"OPTIONS"},
        description="HTTP methods which should be excluded from metrics",
    )
    # These will be new defaults in starlette-exporter 0.18:
    # https://github.com/stephenhillier/starlette_exporter/issues/79
    group_paths: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            If ``True`` (recommended), add request path to metrics literally as described
            in OpenAPI schema, e.g. ``/namespaces/{id}``, without substitution with path real values.

            If ``False``, all real request paths to metrics, e.g. ``/namespaces/123``.
            """,
        ),
    )
    filter_unhandled_paths: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            If ``True``, add metrics for paths only mentioned in OpenAPI schema.

            If ``False``, add all requested paths to metrics.
            """,
        ),
    )

    class Config:
        extra = "allow"



#!/bin/env python3
# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path
from typing import List, Optional

import uvicorn

here = Path(__file__).resolve()


def main(prog_name: Optional[str] = None, args: Optional[List[str]] = None):
    """Run uvicorn and pass the command line arguments to it."""
    if args is None:
        args = sys.argv.copy()
        prog_name = args.pop(0)

    if not prog_name:
        prog_name = os.fspath(here)

    args = args.copy()
    # prepend config path before command line arguments
    args.insert(0, "--factory")
    args.insert(1, "horizon.backend:get_application")

    # call uvicorn
    uvicorn.main.main(args=args, prog_name=prog_name)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

"""
AuthProvider using LDAP for checking user existence and credentials.

Based on `bonsai <https://github.com/noirello/bonsai>`_

Similar to:
* `JupyterHub LDAPAuthenticator <https://github.com/jupyterhub/ldapauthenticator>`_
* `Flask-AppBuilder LDAP integration <https://github.com/dpgaspar/Flask-AppBuilder/blob/master/docs/config.rst>`_, used in Apache Airflow
"""

import logging
from asyncio import TimeoutError
from contextlib import asynccontextmanager
from time import time
from typing import Any, AsyncContextManager, AsyncGenerator, Dict, List, Optional, Tuple

from bonsai import InvalidDN, LDAPClient
from bonsai.asyncio import AIOConnectionPool, AIOLDAPConnection
from bonsai.errors import AuthenticationError, LDAPError
from devtools import pformat
from fastapi import Depends, FastAPI
from typing_extensions import Annotated

from horizon.backend.db.models import User
from horizon.backend.dependencies import Stub
from horizon.backend.providers.auth.base import AuthProvider
from horizon.backend.services import UnitOfWork
from horizon.backend.settings.auth.ldap import LDAPAuthProviderSettings
from horizon.backend.utils.jwt import decode_jwt, sign_jwt
from horizon.commons.exceptions import (
    AuthorizationError,
    EntityNotFoundError,
    ServiceError,
)

log = logging.getLogger(__name__)

LDAPUnrecoverableError = (LDAPError, TimeoutError)


class LDAPAuthProvider(AuthProvider):
    def __init__(
        self,
        auth_settings: Annotated[LDAPAuthProviderSettings, Depends(Stub(LDAPAuthProviderSettings))],
        pool: Annotated[Optional[AIOConnectionPool], Depends(Stub(AIOConnectionPool))],
        unit_of_work: Annotated[UnitOfWork, Depends()],
    ) -> None:
        self._pool: Optional[AIOConnectionPool] = pool
        self._auth_settings: LDAPAuthProviderSettings = auth_settings
        self._uow: UnitOfWork = unit_of_work

    @classmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        auth_settings = LDAPAuthProviderSettings.parse_obj(app.state.settings.auth.dict(exclude={"provider"}))
        log.info("Using %s provider with settings:\n%s", cls.__name__, pformat(auth_settings))
        app.dependency_overrides[AuthProvider] = cls
        app.dependency_overrides[LDAPAuthProviderSettings] = lambda: auth_settings
        pool = cls._create_lookup_pool(auth_settings)
        app.dependency_overrides[AIOConnectionPool] = lambda: pool
        return app

    async def get_current_user(self, access_token: str) -> User:
        if not access_token:
            raise AuthorizationError("Missing auth credentials")

        user_id = self._get_user_id_from_token(access_token)
        user = await self._uow.user.get_by_id(user_id)
        if not user.is_active:
            raise AuthorizationError(f"User {user.username!r} is disabled")
        return user

    async def get_token(
        self,
        grant_type: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> Dict[str, Any]:
        if not login or not password:
            raise AuthorizationError("Missing auth credentials")

        # firstly check if user exists in LDAP and credentials are valid
        username = await self._resolve_username_from_ldap(login, password)

        log.info("Get/create user %r in database", username)
        async with self._uow:
            # and only then create user in database.
            # avoid creating fake users by spamming auth endpoint
            user = await self._uow.user.get_or_create(username=username)

        log.info("User id %r found", user.id)
        if not user.is_active:
            # TODO: check if user is locked in LDAP
            raise AuthorizationError(f"User {username!r} is disabled")

        log.info("Generate access token for user id %r", user.id)
        access_token, expires_at = self._generate_access_token(user)
        return {
            "access_token": access_token,
            "token_type": "bearer",
            "expires_at": expires_at,
        }

    @classmethod
    def _get_lookup_client(cls, settings: LDAPAuthProviderSettings) -> LDAPClient:
        """Create client for lookup queries"""
        client = LDAPClient(str(settings.ldap.url))
        if settings.ldap.lookup.credentials:
            client.set_credentials(
                settings.ldap.auth_mechanism,
                settings.ldap.lookup.credentials.user,
                settings.ldap.lookup.credentials.password.get_secret_value(),
            )
        return client

    @classmethod
    def _create_lookup_pool(cls, settings: LDAPAuthProviderSettings) -> Optional[AIOConnectionPool]:
        """Create and check connection pool for lookup queries"""
        if not settings.ldap.lookup.enabled:
            return None

        client = cls._get_lookup_client(settings)
        if settings.ldap.lookup.check_on_startup:
            try:
                client.connect(timeout=settings.ldap.timeout_seconds)
            except LDAPError as e:
                raise ServiceError("Failed to connect to LDAP") from e

        if not settings.ldap.lookup.pool.enabled:
            return None

        log.debug("Lookup enabled, creating connection pool")
        return AIOConnectionPool(
            client,
            minconn=settings.ldap.lookup.pool.initial,
            maxconn=settings.ldap.lookup.pool.max,
            timeout=settings.ldap.timeout_seconds,
        )

    @asynccontextmanager
    async def _get_lookup_connection(self) -> AsyncGenerator[AIOLDAPConnection, None]:
        """Create connection used for lookup queries"""
        try:
            connect: AsyncContextManager[AIOLDAPConnection]
            if self._pool:
                log.debug("Using lookup pool")
                connect = self._pool.spawn()

            else:
                log.debug("Creating non-pooled connection for lookup")
                client = self._get_lookup_client(self._auth_settings)
                connect = client.connect(is_async=True, timeout=self._auth_settings.ldap.timeout_seconds)

            async with connect as connection:
                try:  # noqa: WPS505
                    yield connection
                except LDAPUnrecoverableError:
                    # avoid returning connection back to the pool
                    # https://bonsai.readthedocs.io/en/latest/advanced.html#connection-timeouts
                    connection.close()
                    raise

        except LDAPUnrecoverableError as e:
            raise ServiceError("Failed to connect to LDAP") from e

    async def _resolve_username_from_ldap(self, login: str, password: str) -> str:
        log.info("Resolve user %r in LDAP", login)
        username = login

        if self._auth_settings.ldap.lookup.enabled:
            log.info("Perform lookup in LDAP")
            dn, username = await self._lookup_user(login)
        else:
            dn = self._auth_settings.ldap.bind_dn_template.format(
                login=login,
                base_dn=self._auth_settings.ldap.base_dn,
                uid_attribute=self._auth_settings.ldap.uid_attribute,
            )

        log.info("Check user credentials %r in LDAP", dn)
        await self._login(dn, password)

        return username

    async def _lookup_user(self, login: str) -> Tuple[str, str]:
        # Reference implementations:
        # https://github.com/dpgaspar/Flask-AppBuilder/blob/2c5763371b81cd679d88b9971ba5d1fc4d71d54b/flask_appbuilder/security/manager.py#L902
        # https://github.com/jupyterhub/ldapauthenticator/blob/main/ldapauthenticator/ldapauthenticator.py
        query = self._auth_settings.ldap.lookup.query_template.format(
            login=login,
            uid_attribute=self._auth_settings.ldap.uid_attribute,
        )
        base_dn = self._auth_settings.ldap.base_dn
        scope = self._auth_settings.ldap.lookup.scope
        log.debug("Lookup query: %r", query)
        log.debug("Base DN: %r", base_dn)
        log.debug("Scope: %r", scope)

        async with self._get_lookup_connection() as connection:
            results = await connection.search(
                base=base_dn,
                scope=scope,
                filter_exp=query,
                attrlist=["*"],
                sizelimit=1,
                timeout=self._auth_settings.ldap.timeout_seconds,
            )

        if not results:
            raise EntityNotFoundError("User", "username", login)

        entry = results[0]
        log.debug("Found entry:\n%s", pformat(entry))
        dn = str(entry["dn"])
        uid = entry[self._auth_settings.ldap.uid_attribute][0]
        return dn, uid

    async def _login(self, login: str, password: str):
        client = LDAPClient(str(self._auth_settings.ldap.url))
        client.set_credentials(self._auth_settings.ldap.auth_mechanism, login, password)
        try:
            connection = client.connect(is_async=True, timeout=self._auth_settings.ldap.timeout_seconds)
            async with connection:
                await connection.whoami()
        except (AuthenticationError, InvalidDN) as e:
            raise AuthorizationError("Wrong credentials") from e
        except LDAPUnrecoverableError as e:
            raise ServiceError("Failed to connect to LDAP") from e

    def _generate_access_token(self, user: User) -> Tuple[str, float]:
        expires_at = time() + self._auth_settings.access_token.expire_seconds
        payload = {
            "user_id": user.id,
            "exp": expires_at,
        }
        access_token = sign_jwt(
            payload,
            self._auth_settings.access_token.secret_key.get_secret_value(),
            self._auth_settings.access_token.security_algorithm,
        )
        return access_token, expires_at

    def _get_user_id_from_token(self, token: str) -> int:
        try:
            payload = decode_jwt(
                token,
                self._auth_settings.access_token.secret_key.get_secret_value(),
                self._auth_settings.access_token.security_algorithm,
            )
            return int(payload["user_id"])
        except (KeyError, TypeError, ValueError) as e:
            raise AuthorizationError("Invalid token") from e



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Increase HWM name limit

Revision ID: 4bc3fffc0209
Revises: dfce9f35c00e
Create Date: 2023-12-23 12:34:38.843814

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4bc3fffc0209"
down_revision = "dfce9f35c00e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "hwm",
        "name",
        existing_type=sa.VARCHAR(length=256),
        type_=sa.String(length=2048),
        existing_nullable=False,
    )
    op.alter_column(
        "hwm_history",
        "name",
        existing_type=sa.VARCHAR(length=256),
        type_=sa.String(length=2048),
        existing_nullable=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "hwm_history",
        "name",
        existing_type=sa.String(length=2048),
        type_=sa.VARCHAR(length=256),
        existing_nullable=False,
    )
    op.alter_column(
        "hwm",
        "name",
        existing_type=sa.String(length=2048),
        type_=sa.VARCHAR(length=256),
        existing_nullable=False,
    )
    # ### end Alembic commands ###



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add namespace_user table
Revision ID: 4c60578f3f1d
Revises: 09be8fd79dbc
Create Date: 2024-03-01 11:30:36.778975

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4c60578f3f1d"
down_revision = "09be8fd79dbc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "namespace_user",
        sa.Column("namespace_id", sa.BigInteger(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("role", sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(
            ["namespace_id"],
            ["namespace.id"],
            name=op.f("fk__namespace_user__namespace_id__namespace"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk__namespace_user__user_id__user"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("namespace_id", "user_id", name=op.f("pk__namespace_user")),
    )
    # Any user who has changed a namespace at least once should be a maintainer
    op.execute(
        """
        INSERT INTO namespace_user (namespace_id, user_id, role)
            SELECT DISTINCT
                namespace_history.namespace_id,
                namespace_history.changed_by_user_id,
                'MAINTAINER' AS role
            FROM namespace_history
            JOIN namespace ON namespace_history.namespace_id = namespace.id
            WHERE
                namespace_history.changed_by_user_id IS NOT NULL
            AND namespace_history.changed_by_user_id != namespace.owner_id
        ON CONFLICT (namespace_id, user_id) DO NOTHING
        """,
    )
    # Any user who has changed any HWM in namespace at least once should be a developer
    op.execute(
        """
        INSERT INTO namespace_user (namespace_id, user_id, role)
            SELECT DISTINCT
                hwm_history.namespace_id,
                hwm_history.changed_by_user_id,
                'DEVELOPER' AS role
            FROM hwm_history
            JOIN namespace ON hwm_history.namespace_id = namespace.id
            WHERE
                hwm_history.changed_by_user_id IS NOT NULL
            AND hwm_history.changed_by_user_id != namespace.owner_id
        ON CONFLICT (namespace_id, user_id) DO NOTHING
        """,
    )


def downgrade() -> None:
    op.drop_table("namespace_user")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pydantic import BaseModel
from typing_extensions import Literal


class PingResponse(BaseModel):
    """Ping result"""

    status: Literal["ok"] = "ok"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import (
    APIErrorResponse,
    get_error_responses,
    get_response_for_exception,
    get_response_for_status_code,
)

__all__ = [
    "APIErrorResponse",
    "BaseErrorSchema",
    "get_error_responses",
    "get_response_for_exception",
    "get_response_for_status_code",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

"""
AuthProvider using LDAP, but
"""

import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional

from bonsai.asyncio import AIOConnectionPool
from devtools import pformat
from fastapi import Depends, FastAPI
from passlib.ifc import PasswordHash
from passlib.registry import get_crypt_handler
from typing_extensions import Annotated

from horizon.backend.dependencies import Stub
from horizon.backend.providers.auth.base import AuthProvider
from horizon.backend.providers.auth.ldap import LDAPAuthProvider
from horizon.backend.services import UnitOfWork
from horizon.backend.settings.auth.cached_ldap import CachedLDAPAuthProviderSettings
from horizon.commons.exceptions import AuthorizationError

log = logging.getLogger(__name__)


class CachedLDAPAuthProvider(LDAPAuthProvider):
    def __init__(
        self,
        auth_settings: Annotated[CachedLDAPAuthProviderSettings, Depends(Stub(CachedLDAPAuthProviderSettings))],
        pool: Annotated[Optional[AIOConnectionPool], Depends(Stub(AIOConnectionPool))],
        unit_of_work: Annotated[UnitOfWork, Depends()],
    ) -> None:
        self._pool: Optional[AIOConnectionPool] = pool
        self._auth_settings: CachedLDAPAuthProviderSettings = auth_settings
        self._uow: UnitOfWork = unit_of_work

    @classmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        auth_settings = CachedLDAPAuthProviderSettings.parse_obj(app.state.settings.auth.dict(exclude={"provider"}))
        log.info("Using %s provider with settings:\n%s", cls.__name__, pformat(auth_settings))
        app.dependency_overrides[AuthProvider] = cls
        app.dependency_overrides[CachedLDAPAuthProviderSettings] = lambda: auth_settings
        pool = cls._create_lookup_pool(auth_settings)
        app.dependency_overrides[AIOConnectionPool] = lambda: pool
        return app

    async def get_token(
        self,
        grant_type: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> Dict[str, Any]:
        if not login or not password:
            raise AuthorizationError("Missing auth credentials")

        # firstly check if user credentials already exists in cache
        from_cache = True
        username = await self._resolve_username_from_credentials_cache(login, password)
        if not username:
            from_cache = False
            username = await self._resolve_username_from_ldap(login, password)

        log.info("Get/create user %r in database", username)
        async with self._uow:
            # and only then create user in database.
            # avoid creating fake users by spamming auth endpoint
            user = await self._uow.user.get_or_create(username=username)

            log.info("User id %r found", user.id)
            if not user.is_active:
                # TODO: check if user is locked in LDAP
                raise AuthorizationError(f"User {username!r} is disabled")

            if not from_cache:
                # updating cache without checking user in LDAP means cache item will never expire,
                # and we will never check if used is valid
                log.info("Update credentials cache for user id %r", user.id)
                await self._update_credentials_cache(user_id=user.id, login=login, password=password)

        log.info("Generate access token for user id %r", user.id)
        access_token, expires_at = self._generate_access_token(user)
        return {
            "access_token": access_token,
            "token_type": "bearer",
            "expires_at": expires_at,
        }

    def _get_hasher(self) -> PasswordHash:
        hash_settings = self._auth_settings.cache.password_hash
        handler = get_crypt_handler(hash_settings.algorithm)
        return handler.using(**hash_settings.options)

    async def _resolve_username_from_credentials_cache(self, login: str, password: str) -> Optional[str]:
        log.info("Perform lookup in credentials cache")
        user_cache = await self._uow.credentials_cache.get_by_login(login)
        if not user_cache:
            log.info("User %r not found in cache", login)
            return None

        log.debug("Found item %r", user_cache)
        expiration_date = user_cache.updated_at + timedelta(seconds=self._auth_settings.cache.expire_seconds)
        if expiration_date < datetime.now(tz=timezone.utc):
            log.info("Cache item expired")
            return None

        hasher = self._get_hasher()
        if not hasher.verify(password, user_cache.password_hash):
            raise AuthorizationError("Wrong credentials")

        log.info("Credentials match the cache")
        return user_cache.user.username

    async def _update_credentials_cache(self, user_id: int, login: str, password: str) -> None:
        # this is not a dedicated method of repository because we need hashing settings to generate password hash,
        # and generating new hash every time is expensive
        hasher = self._get_hasher()
        user_cache = await self._uow.credentials_cache.get_by_login(login=login)

        data: Dict[str, Any] = {}
        if not user_cache or user_cache.user_id != user_id:
            data["user_id"] = user_id
        if not user_cache or not hasher.verify(password, user_cache.password_hash):
            data["password_hash"] = hasher.hash(password)

        # even if nothing is changed in credentials cache, we do perform an update to sync `updated_at` column
        await self._uow.credentials_cache.create_or_update(login=login, data=data)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from fastapi import FastAPI
from starlette.datastructures import MutableHeaders

from horizon.backend.settings.server.application_version import (
    ApplicationVersionSettings,
)

if TYPE_CHECKING:
    from starlette.types import ASGIApp, Message, Receive, Scope, Send


@dataclass
class ApplicationVersionMiddleware:
    app: ASGIApp
    version: str
    header_name: str = "X-Application-Version"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in {"http", "websocket"}:
            # lifespan or other type of request
            await self.app(scope, receive, send)
            return

        async def modify_response_headers(message: Message) -> None:  # noqa: WPS430
            if message["type"] == "http.response.start":
                headers = MutableHeaders(scope=message)
                headers.append(self.header_name, self.version)

            await send(message)

        await self.app(scope, receive, modify_response_headers)


def apply_application_version_middleware(app: FastAPI, settings: ApplicationVersionSettings) -> FastAPI:
    """Add X-Application-Version middleware to the application."""
    if not settings.enabled:
        return app

    app.add_middleware(
        ApplicationVersionMiddleware,
        version=app.version,
        **settings.dict(exclude={"enabled"}),
    )
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import re
from typing import List, Sequence

from sqlalchemy import SQLColumnExpression, delete, select
from sqlalchemy.exc import IntegrityError

from horizon.backend.db.models import HWM, HWMHistory, User
from horizon.backend.db.repositories.base import Repository
from horizon.commons.dto.pagination import Pagination
from horizon.commons.exceptions.entity import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
)


class HWMRepository(Repository[HWM]):
    async def paginate(
        self,
        namespace_id: int,
        page: int,
        page_size: int,
        name: str | None = None,
    ) -> Pagination[HWM]:
        where: list[SQLColumnExpression] = [
            HWM.namespace_id == namespace_id,
        ]
        if name:
            where.append(HWM.name == name)

        return await self._paginate(
            where=where,
            order_by=[HWM.name],
            page=page,
            page_size=page_size,
        )

    async def count(self) -> int:
        return await self._count()

    async def get(
        self,
        hwm_id: int,
    ) -> HWM:
        result = await self._get(
            HWM.id == hwm_id,
        )
        if not result:
            raise EntityNotFoundError("HWM", "id", hwm_id)
        return result

    async def create(
        self,
        data: dict,
        user: User,
    ) -> HWM:
        try:
            result = await self._create(data={**data, "changed_by_user_id": user.id})
            await self._session.flush()
            return result
        except IntegrityError as e:
            constraint = e.__cause__.__cause__.constraint_name  # type: ignore[union-attr]

            if constraint == "fk__hwm__namespace_id__namespace":
                raise EntityNotFoundError("Namespace", "id", data["namespace_id"]) from e

            if constraint == "hwm_name_unique_per_namespace":
                raise EntityAlreadyExistsError("HWM", "name", data["name"]) from e

            raise

    async def update(
        self,
        hwm_id: int,
        changes: dict,
        user: User,
    ) -> HWM:
        try:
            result = await self._update(
                where=[HWM.id == hwm_id],
                changes={**changes, "changed_by_user_id": user.id},
            )
            if result is None:
                raise EntityNotFoundError("HWM", "id", hwm_id)

            await self._session.flush()
            await self._session.refresh(result)
            return result
        except IntegrityError as e:
            raise EntityAlreadyExistsError("HWM", "name", changes["name"]) from e

    async def delete(
        self,
        hwm_id: int,
    ) -> HWM:
        hwm = await self.get(hwm_id)
        await self._session.delete(hwm)
        await self._session.flush()
        return hwm

    async def bulk_delete(self, namespace_id: int, hwm_ids: List[int]) -> Sequence[HWM]:
        result = await self._session.execute(select(HWM).where(HWM.id.in_(hwm_ids), HWM.namespace_id == namespace_id))
        hwms_to_delete = result.scalars().all()

        if hwms_to_delete:
            await self._session.execute(delete(HWM).where(HWM.id.in_([hwm.id for hwm in hwms_to_delete])))

        return hwms_to_delete

    async def bulk_copy(
        self,
        source_namespace_id: int,
        target_namespace_id: int,
        hwm_ids: list[int],
        with_history: bool,
    ) -> Sequence[HWM]:
        result = await self._session.execute(
            select(HWM).where(HWM.id.in_(hwm_ids), HWM.namespace_id == source_namespace_id),
        )
        hwms = result.scalars().all()
        copied_hwms = []

        for hwm in hwms:
            new_hwm = HWM(
                **hwm.to_dict(exclude={"id", "namespace_id"}),
                namespace_id=target_namespace_id,
            )
            self._session.add(new_hwm)
            copied_hwms.append(new_hwm)

        try:
            await self._session.flush()
        except IntegrityError as e:
            constraint = e.__cause__.__cause__.constraint_name  # type: ignore[union-attr]
            if constraint == "hwm_name_unique_per_namespace":
                hwm_name = re.search(r"Key \(namespace_id, name\)=\(\d+, (.+)\) already exists.", e.__cause__.__cause__.detail).group(1)  # type: ignore[union-attr]
                raise EntityAlreadyExistsError("HWM", "name", hwm_name) from e

        if with_history:
            for original_hwm, copied_hwm in zip(hwms, copied_hwms):
                history = await self._session.execute(
                    select(HWMHistory).where(HWMHistory.hwm_id == original_hwm.id),
                )
                history = history.scalars().all()  # type: ignore
                for record in history:
                    new_history_record = HWMHistory(
                        **record.to_dict(exclude={"id", "hwm_id"}),
                        hwm_id=copied_hwm.id,
                    )
                    self._session.add(new_history_record)

            await self._session.flush()

        return copied_hwms



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version
from pydantic import root_validator

from horizon.commons.dto import Unset
from horizon.commons.schemas.v1.pagination import PaginateQueryV1

MAX_NAME_LENGTH = 256


class NamespaceUserRole(str, Enum):  # noqa: WPS60
    DEVELOPER = "DEVELOPER"
    MAINTAINER = "MAINTAINER"
    OWNER = "OWNER"


class NamespaceResponseV1(BaseModel):
    """Namespace response."""

    id: int = Field(description="Namespace id")
    name: str = Field(description="Namespace name, unique in the entire database")
    description: str = Field(description="Namespace description")
    owned_by: str = Field(description="The namespace owner")
    changed_at: datetime = Field(description="Timestamp of last change of the namespace data")
    changed_by: Optional[str] = Field(default=None, description="Latest user who changed the namespace data")

    class Config:
        if pydantic_version >= "2":
            from_attributes = True
        else:
            orm_mode = True


class NamespacePaginateQueryV1(PaginateQueryV1):
    """Query params for namespace pagination request."""

    name: Optional[str] = Field(default=None, min_length=1, max_length=MAX_NAME_LENGTH)

    # more arguments can be added in future


class NamespaceCreateRequestV1(BaseModel):
    """Request body for namespace creation request."""

    name: str = Field(min_length=1, max_length=MAX_NAME_LENGTH)
    description: str = ""


class NamespaceUpdateRequestV1(BaseModel):
    """Request body for namespace update request.

    If field value is not set, it will not be updated.
    """

    name: Union[str, Unset] = Field(default=Unset(), min_length=1, max_length=MAX_NAME_LENGTH)
    description: Union[str, Unset] = Unset()

    class Config:
        arbitrary_types_allowed = True

    @root_validator(skip_on_failure=True)
    def _any_field_set(cls, values):
        """Validate that at least one field is set."""
        values_set = {k for k, v in values.items() if not isinstance(v, Unset)}
        if not values_set:
            raise ValueError("At least one field must be set.")
        return values



# SPDX-FileCopyrightText: 2023 Andrey Tikhonov (@Tishka17)
# SPDX-License-Identifier: Apache-2.0


# Implementation copied from:
# https://github.com/Tishka17/deseos17/blob/master/src/deseos17/presentation/web_api/dependencies/depends_stub.py
class Stub:
    """
    This class is used to prevent fastapi from digging into
    real dependencies attributes detecting them as request data

    So instead of
    ``value: Annotated[MyDependency, Depends()]``
    Write
    ``value: Annotated[MyDependency, Depends(Stub(MyDependency))]``

    And then you can declare how to create it:
    ``app.dependency_overrides[MyDependency] = my_dependency_factory``

    """

    def __init__(self, dependency: type, **kwargs):
        self._dependency = dependency
        self._kwargs = kwargs

    def __call__(self):
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        if isinstance(other, Stub):
            return (
                self._dependency == other._dependency and self._kwargs == other._kwargs  # noqa: WPS437  # noqa: WPS437
            )
        if not self._kwargs:
            return self._dependency == other
        return False

    def __hash__(self):
        if not self._kwargs:
            return hash(self._dependency)
        serial = (
            self._dependency,
            *self._kwargs.items(),
        )
        return hash(serial)

    def __repr__(self):
        args = "".join(f", {k}={v!r}" for k, v in self._kwargs.items())
        class_name = self.__class__.__name__
        dependency_class = self._dependency.__name__
        return f"{class_name}({dependency_class}{args})"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from sqlalchemy import BigInteger, Boolean, String
from sqlalchemy.orm import Mapped, mapped_column

from horizon.backend.db.mixins.timestamp import TimestampMixin
from horizon.backend.db.models.base import Base


class User(Base, TimestampMixin):
    __tablename__ = "user"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    username: Mapped[str] = mapped_column(
        String(256),  # noqa: WPS432
        nullable=False,
        unique=True,
        index=True,
    )
    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    is_admin: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Generic, TypeVar

from pydantic import BaseModel
from pydantic import __version__ as pydantic_version

if pydantic_version >= "2":
    from pydantic import BaseModel as GenericModel
else:
    from pydantic.generics import GenericModel  # type: ignore[no-redef] # noqa: WPS440

from horizon.commons.dto.unset import Unset


class BaseErrorSchema(BaseModel):
    code: str
    message: str
    details: Any


ErrorModel = TypeVar("ErrorModel", bound=BaseErrorSchema)


class APIErrorSchema(GenericModel, Generic[ErrorModel]):
    error: ErrorModel

    if pydantic_version < "2":

        class Config:
            # https://github.com/pydantic/pydantic/issues/2277
            json_encoders = {
                Unset: str,
            }



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Dict, cast

from sqlalchemy import SQLColumnExpression, delete, select, update
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.exc import IntegrityError

from horizon.backend.db.models import (
    Namespace,
    NamespaceUser,
    NamespaceUserRoleInt,
    User,
)
from horizon.backend.db.repositories.base import Repository
from horizon.commons.dto import Pagination
from horizon.commons.exceptions import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
    PermissionDeniedError,
)


class NamespaceRepository(Repository[Namespace]):
    async def paginate(
        self,
        page: int,
        page_size: int,
        name: str | None = None,
    ) -> Pagination[Namespace]:
        where: list[SQLColumnExpression] = [Namespace.name == name] if name else []

        return await self._paginate(
            where=where,
            order_by=[Namespace.name],
            page=page,
            page_size=page_size,
        )

    async def count(self) -> int:
        return await self._count()

    async def get(
        self,
        namespace_id: int,
    ) -> Namespace:
        result = await self._get(
            Namespace.id == namespace_id,
        )
        if not result:
            raise EntityNotFoundError("Namespace", "id", namespace_id)
        return result

    async def create(
        self,
        name: str,
        description: str,
        user: User,
    ) -> Namespace:
        try:
            result = await self._create(
                data={
                    "name": name,
                    "description": description,
                    "changed_by_user_id": user.id,
                    "owner_id": user.id,
                },
            )
            await self._session.flush()
            return result
        except IntegrityError as e:
            raise EntityAlreadyExistsError("Namespace", "name", name) from e

    async def update(
        self,
        namespace_id: int,
        changes: dict,
        user: User,
    ) -> Namespace:
        try:
            result = await self._update(
                where=[Namespace.id == namespace_id],
                changes={**changes, "changed_by_user_id": user.id},
            )
            if result is None:
                raise EntityNotFoundError("Namespace", "id", namespace_id)

            await self._session.flush()
            return result
        except IntegrityError as e:
            raise EntityAlreadyExistsError(
                "Namespace",
                "name",
                changes.get("name"),
            ) from e

    async def delete(
        self,
        namespace_id: int,
        user: User,
    ) -> Namespace:
        namespace = await self.get(namespace_id)
        await self._session.delete(namespace)
        await self._session.flush()
        return namespace

    async def check_user_permission(self, user_id: int, namespace_id: int, required_role: NamespaceUserRoleInt) -> None:

        # if the user is a SUPERADMIN, they inherently have all permissions
        superadmin_user = await self._session.execute(select(User.is_admin).where(User.id == user_id))
        if superadmin_user.scalar_one_or_none():
            return

        owner_check = await self._session.execute(select(Namespace.owner_id).where(Namespace.id == namespace_id))
        owner_id = owner_check.scalar_one_or_none()
        if owner_id is None:
            raise EntityNotFoundError("Namespace", "id", namespace_id)

        if owner_id == user_id:
            user_role = NamespaceUserRoleInt.OWNER
        else:
            role_result = await self._session.execute(
                select(NamespaceUser.role).where(
                    NamespaceUser.namespace_id == namespace_id,
                    NamespaceUser.user_id == user_id,
                ),
            )
            user_role_value = role_result.scalars().first()
            user_role = NamespaceUserRoleInt[user_role_value] if user_role_value else NamespaceUserRoleInt.GUEST

        if user_role < required_role:
            raise PermissionDeniedError(required_role.name, user_role.name)

    async def get_namespace_users_permissions(self, namespace_id: int) -> Dict[User, NamespaceUserRoleInt]:
        permissions_dict = {}

        namespace = await self.get(namespace_id)
        owner = await self._session.get(User, namespace.owner_id)
        owner = cast(User, owner)
        permissions_dict[owner] = NamespaceUserRoleInt.OWNER

        query = (
            select(User, NamespaceUser.role)
            .join(NamespaceUser, User.id == NamespaceUser.user_id)
            .where(NamespaceUser.namespace_id == namespace_id)
            .order_by(NamespaceUser.role.desc(), User.username)
        )
        result = await self._session.execute(query)
        for user, role in result.fetchall():
            permissions_dict[user] = NamespaceUserRoleInt[role]

        return permissions_dict

    async def set_new_owner(self, namespace_id: int, new_owner_id: int) -> None:
        await self._session.execute(update(Namespace).where(Namespace.id == namespace_id).values(owner_id=new_owner_id))
        await self._session.execute(
            delete(NamespaceUser).where(
                NamespaceUser.namespace_id == namespace_id,
                NamespaceUser.user_id == new_owner_id,
            ),
        )

    async def update_permission(self, namespace_id: int, user_id: int, role: NamespaceUserRoleInt) -> None:
        # create an "upsert" statement using PostgreSQL's ON CONFLICT syntax.
        stmt = pg_insert(NamespaceUser).values(namespace_id=namespace_id, user_id=user_id, role=role.name)
        stmt = stmt.on_conflict_do_update(
            index_elements=[NamespaceUser.namespace_id, NamespaceUser.user_id],
            set_={"role": role.name},
        )
        await self._session.execute(stmt)

    async def delete_permission(self, namespace_id: int, user_id: int) -> None:
        await self._session.execute(
            delete(NamespaceUser).where(NamespaceUser.namespace_id == namespace_id, NamespaceUser.user_id == user_id),
        )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import http
import logging
import pprint
import warnings
from typing import Any, Generic, Optional, Tuple, TypeVar
from urllib.parse import urlparse

from pydantic import AnyHttpUrl, BaseModel, PrivateAttr, ValidationError
from pydantic import __version__ as pydantic_version
from pydantic import parse_obj_as, validator

if pydantic_version >= "2":
    from pydantic import BaseModel as GenericModel
else:
    from pydantic.generics import GenericModel  # type: ignore[no-redef] # noqa: WPS440
from typing_extensions import Protocol

import horizon
from horizon.client.auth.base import BaseAuth
from horizon.commons.errors import get_response_for_status_code
from horizon.commons.errors.base import APIErrorSchema

logger = logging.getLogger(__name__)

SessionClass = TypeVar("SessionClass")
ResponseSchema = TypeVar("ResponseSchema", bound=BaseModel)


class BaseResponse(Protocol):
    """Response-like object. Same interface is shared between requests.Response and httpx.Response"""

    @property
    def status_code(self) -> int: ...  # noqa: WPS473

    @property
    def content(self) -> Any: ...

    def json(self) -> Any: ...

    def raise_for_status(self) -> None: ...

    @property
    def headers(self) -> dict[str, str]: ...


class BaseClient(GenericModel, Generic[SessionClass]):
    """Base Horizon client implementation, designed to be subclassed"""

    base_url: AnyHttpUrl
    auth: BaseAuth
    session: Optional[SessionClass] = None

    _backend_version_tuple: Tuple[int, ...] = PrivateAttr(default_factory=tuple)

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def session_class(cls) -> type[SessionClass]:
        # Get `Session` from `SyncClient(BaseClient[Session])`
        if pydantic_version >= "2":
            return cls.model_fields["session"].annotation.__args__[0]  # type: ignore[union-attr]
        return cls.__bases__[0].__annotations__["session"].__args__[0]

    @validator("base_url")
    def _validate_url(cls, value: AnyHttpUrl):
        """``http://localhost:8000/`` -> ``http://localhost:8000``"""
        if value.path:
            return urlparse(str(value))._replace(path=value.path.rstrip("/")).geturl()  # noqa: WPS437
        return value

    @validator("session", always=True)
    def _default_session(cls, session: SessionClass | None):
        """If session is not passed, create it automatically"""
        if session:
            return session
        RealSessionClass = cls.session_class()  # noqa: N806
        return RealSessionClass()

    @validator("session", always=True)
    def _patch_session(cls, session: SessionClass, values: dict):
        """Patch session for chosen auth method, if required"""
        auth: BaseAuth = values.get("auth")  # type: ignore[assignment]
        return auth.patch_session(session)

    def _parse_body(self, body: dict, response_class: type[ResponseSchema]) -> ResponseSchema:
        try:
            return parse_obj_as(response_class, body)
        except ValidationError as e:  # noqa: WPS329
            # Response does not match expected schema. Probably API was changed.
            # ValidationError does not contain body, so we attaching it to response.
            raise e from ValueError(body)

    def _handle_backend_version(self, backend_version: str | None):
        if self._backend_version_tuple or not backend_version:
            return

        self._backend_version_tuple = tuple(map(int, backend_version.split(".")))  # noqa: WPS601
        if self._backend_version_tuple > horizon.__version_tuple__:
            message = (
                f"Horizon client version {horizon.__version__!r} does not match backend version {backend_version!r}. "
                "Please upgrade."
            )
            warnings.warn(message, UserWarning, stacklevel=5)

    def _handle_response(  # noqa: WPS238, WPS231
        self,
        response: BaseResponse,
        response_class: type[ResponseSchema] | None,
    ) -> ResponseSchema | None:
        """Convert Response object to expected response class, or raise an exception matching the status code"""
        request_id: str | None = response.headers.get("X-Request-ID", None)
        if request_id:
            logger.debug("Request ID: %r", request_id)

        backend_version = response.headers.get("X-Application-Version", None)
        self._handle_backend_version(backend_version)

        if response.status_code == http.HTTPStatus.NO_CONTENT.value:
            return None

        if response.status_code < http.HTTPStatus.BAD_REQUEST.value and response_class:
            return self._parse_body(response.json(), response_class)

        # raise_for_exception will definitely raise something, but mypy does not know that
        # so we create some exception to be bypass it
        http_exception: Exception = AssertionError("If you see this message, something went wrong")
        try:
            response.raise_for_status()
        except Exception as e:
            http_exception = e

        if request_id and hasattr(http_exception, "add_note"):
            # add_note is only available in Python 3.11+
            # https://docs.python.org/3/library/exceptions.html#BaseException.add_note
            http_exception.add_note(f"Request ID: {request_id!r}")

        try:
            body = response.json()
            if hasattr(http_exception, "add_note"):
                http_exception.add_note(f"Response body:\n{pprint.pformat(body)}")  # noqa: WPS237

        except Exception as format_err:  # noqa: WPS329
            if hasattr(http_exception, "add_note"):
                http_exception.add_note(f"Response body:\n{response.content!r}")
            else:
                logger.error("Response body:\n%r", response.content)

            raise format_err from http_exception

        error_response = get_response_for_status_code(response.status_code)
        if not error_response:
            # cannot handle this status code
            raise http_exception

        try:
            error_value = parse_obj_as(APIErrorSchema[error_response.schema], body)  # type: ignore[name-defined]
        except ValidationError:  # noqa: WPS329
            # Something wrong with API response, probably wrong URL
            raise http_exception

        get_exception = getattr(error_value.error, "to_exception", None)
        if get_exception:
            raise get_exception() from http_exception

        raise http_exception



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add credentials_cache table

Revision ID: dfce9f35c00e
Revises: 6b9001985cd2
Create Date: 2023-11-24 12:45:37.006395

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "dfce9f35c00e"
down_revision = "6b9001985cd2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "credentials_cache",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("login", sa.String(length=256), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("password_hash", sa.Text(), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk__credentials_cache__user_id__user"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__credentials_cache")),
    )
    op.create_index(op.f("ix__credentials_cache__login"), "credentials_cache", ["login"], unique=True)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix__credentials_cache__login"), table_name="credentials_cache")
    op.drop_table("credentials_cache")
    # ### end Alembic commands ###



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0


from horizon.commons.exceptions.base import ApplicationError


class ServiceError(ApplicationError):
    """Service used by application have not responded properly.

    Examples
    --------

    >>> from horizon.commons.exceptions import ServiceError
    >>> raise ServiceError("Some server response is invalid")
    Traceback (most recent call last):
    horizon.commons.exceptions.service.ServiceError: Some server response is invalid
    """

    def __init__(self, message: str) -> None:
        self._message = message

    @property
    def message(self) -> str:
        return self._message

    @property
    def details(self) -> None:
        pass



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import FastAPI
from uuid6 import uuid7

from horizon.backend.settings.server import RequestIDSettings


def apply_request_id_middleware(app: FastAPI, settings: RequestIDSettings) -> FastAPI:
    """Add X-Request-ID middleware to the application."""
    if not settings.enabled:
        return app

    app.add_middleware(
        CorrelationIdMiddleware,
        generator=lambda: uuid7().hex,
        **settings.dict(exclude={"enabled"}),
    )
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap

from pydantic import BaseModel, Field


class RequestIDSettings(BaseModel):
    """X-Request-ID Middleware Settings.

    See `asgi-correlation-id <https://github.com/snok/asgi-correlation-id>`_ documentation.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__REQUEST_ID__ENABLED=True
        HORIZON__SERVER__REQUEST_ID__UPDATE_REQUEST_HEADER=True
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable middleware")
    header_name: str = Field(
        default="X-Request-ID",
        description="Name of response header which is filled up with request ID value",
    )
    update_request_header: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            If ``False``, bypass header value from request to response as-is.

            If ``True``, always set new value of specific header.
            """,
        ),
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version

from horizon.commons.schemas.v1.pagination import PaginateQueryV1


class HWMHistoryPaginateQueryV1(PaginateQueryV1):
    """Query params for HWM pagination request."""

    hwm_id: int = Field(description="HWM id")

    # more arguments can be added in future


class HWMHistoryResponseV1(BaseModel):
    """HWM history response."""

    id: int = Field(description="HWM history item id")
    hwm_id: int = Field(description="HWM id")
    namespace_id: int = Field(description="Namespace id HWM is bound to")
    name: str = Field(description="HWM name, unique in the namespace")
    description: str = Field(description="HWM description")
    type: str = Field(description="HWM type, any non-empty string")
    value: Any = Field(description="HWM value, any JSON serializable value")
    entity: Optional[str] = Field(default=None, description="Name of entity associated with the HWM. Can be any string")
    expression: Optional[str] = Field(
        default=None,
        description="Expression used to calculate HWM value. Can be any string",
    )
    action: str = Field(description="Action performed on the HWM record")
    changed_at: datetime = Field(description="Timestamp of a change of the HWM data")
    changed_by: Optional[str] = Field(default=None, description="User who changed the HWM data")

    class Config:
        if pydantic_version >= "2":
            from_attributes = True
        else:
            orm_mode = True



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import APIRouter, Depends
from typing_extensions import Annotated

from horizon.backend.services import UnitOfWork, current_user
from horizon.commons.errors import get_error_responses
from horizon.commons.schemas.v1 import (
    NamespaceHistoryPaginateQueryV1,
    NamespaceHistoryResponseV1,
    PageResponseV1,
)

router = APIRouter(prefix="/namespace-history", tags=["Namespace History"], responses=get_error_responses())


@router.get(
    "/",
    summary="Paginate namespace history",
    dependencies=[Depends(current_user)],
)
async def paginate_hwm_history(
    pagination_args: Annotated[NamespaceHistoryPaginateQueryV1, Depends()],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> PageResponseV1[NamespaceHistoryResponseV1]:
    pagination = await unit_of_work.namespace_history.paginate(**pagination_args.dict())
    return PageResponseV1[NamespaceHistoryResponseV1].from_pagination(pagination)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap

from pydantic import BaseModel, Field

from horizon.backend.settings.server.application_version import (
    ApplicationVersionSettings,
)
from horizon.backend.settings.server.cors import CORSSettings
from horizon.backend.settings.server.log import LoggingSettings
from horizon.backend.settings.server.monitoring import MonitoringSettings
from horizon.backend.settings.server.openapi import OpenAPISettings
from horizon.backend.settings.server.request_id import RequestIDSettings
from horizon.backend.settings.server.static_files import StaticFilesSettings


class ServerSettings(BaseModel):
    """Backend server settings.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__DEBUG=True
        HORIZON__SERVER__LOGGING__PRESET=colored
        HORIZON__SERVER__MONITORING__ENABLED=True
        HORIZON__SERVER__CORS__ENABLED=True
        HORIZON__SERVER__REQUEST_ID__ENABLED=True
        HORIZON__SERVER__OPENAPI__ENABLED=True
        HORIZON__SERVER__OPENAPI__SWAGGER__ENABLED=True
        HORIZON__SERVER__OPENAPI__REDOC__ENABLED=True
    """

    debug: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            :ref:`Enable debug output in responses <backend-configuration-debug>`.
            Do not use this on production!
            """,
        ),
    )
    logging: LoggingSettings = Field(
        default_factory=LoggingSettings,
        description=":ref:`Logging settings <backend-configuration-logging>`",
    )
    cors: CORSSettings = Field(
        default_factory=CORSSettings,
        description=":ref:`CORS settings <backend-configuration-cors>`",
    )
    monitoring: MonitoringSettings = Field(
        default_factory=MonitoringSettings,
        description=":ref:`Monitoring settings <backend-configuration-monitoring>`",
    )
    request_id: RequestIDSettings = Field(
        default_factory=RequestIDSettings,
        description=":ref:`RequestID settings <backend-configuration-debug>`",
    )
    application_version: ApplicationVersionSettings = Field(
        default_factory=ApplicationVersionSettings,
        description=":ref:`Application version settings <backend-configuration-debug>`",
    )
    static_files: StaticFilesSettings = Field(
        default_factory=StaticFilesSettings,
        description=":ref:`Static files settings <backend-configuration-static-files>`",
    )
    openapi: OpenAPISettings = Field(
        default_factory=OpenAPISettings,
        description=":ref:`OpenAPI.json settings <backend-configuration-openapi>`",
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import ClassVar, Optional


class Unset:  # noqa: WPS600
    _instance: ClassVar[Optional[Unset]] = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __len__(self):
        # handle ``min_length`` validator
        return 1

    def __repr__(self):
        return "Unset()"

    def __str__(self):
        return "<unset>"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime, timezone
from functools import partial

from sqlalchemy import BigInteger, DateTime, ForeignKey, func
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Mapped, mapped_column, relationship

from horizon.backend.db.models.user import User


class ChangedByMixin:
    changed_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=partial(datetime.now, tz=timezone.utc),
        nullable=False,
    )
    changed_by_user_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("user.id", ondelete="SET NULL"),
        nullable=True,
    )

    @declared_attr
    def changed_by_user(cls):
        return relationship(User, foreign_keys=[cls.changed_by_user_id], lazy="selectin")

    @declared_attr
    def changed_by(cls):
        return association_proxy("changed_by_user", "username")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

BaseSettings: type
try:
    from pydantic import BaseSettings  # type: ignore[no-redef] # noqa: WPS440
except ImportError:
    from pydantic_settings import BaseSettings  # type: ignore[no-redef] # noqa: WPS440

from horizon.backend.settings.auth import AuthSettings
from horizon.backend.settings.database import DatabaseSettings
from horizon.backend.settings.server import ServerSettings


class Settings(BaseSettings):
    """Horizon backend settings.

    Backend can be configured in 2 ways:

    * By explicitly passing ``settings`` object as an argument to :obj:`application_factory <horizon.backend.main.application_factory>`
    * By setting up environment variables matching a specific key.

        All environment variable names are written in uppercase and should be prefixed with ``HORIZON__``.
        Nested items are delimited with ``__``.

    More details can be found in `Pydantic documentation <https://docs.pydantic.dev/latest/concepts/pydantic_settings/>`_.

    Examples
    --------

    .. code-block:: bash

        # same as settings.database.url = "postgresql+asyncpg://postgres:postgres@localhost:5432/horizon"
        HORIZON__DATABASE__URL=postgresql+asyncpg://postgres:postgres@localhost:5432/horizon

        # same as settings.server.debug = True
        HORIZON__SERVER__DEBUG=True

        # same as settings.auth.provider = horizon.backend.providers.auth.dummy.DummyAuthProvider
        HORIZON__AUTH__PROVIDER=horizon.backend.providers.auth.dummy.DummyAuthProvider
    """

    database: DatabaseSettings = Field(description=":ref:`Database settings <backend-configuration-database>`")
    server: ServerSettings = Field(
        default_factory=ServerSettings,
        description=":ref:`Server settings <backend-configuration>`",
    )
    auth: AuthSettings = Field(
        default_factory=AuthSettings,
        description=":ref:`Auth setting <backend-auth-providers>`",
    )

    class Config:
        env_prefix = "HORIZON__"
        env_nested_delimiter = "__"
